unit Unit1;

interface

uses
  SysUtils, Types, Classes, Variants, QTypes, opbitmap, opbitmapformats, XPMime, QGraphics, QControls, QForms,
  QDialogs, QStdCtrls, QExtCtrls, QComCtrls;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Button1: TButton;
    ScrollBox1: TScrollBox;
    Image1: TImage;
    OpenDialog1: TOpenDialog;
    ProgressBar1: TProgressBar;
    Button2: TButton;
    SaveDialog1: TSaveDialog;
    cbPixelFormat: TComboBox;
    Button3: TButton;
    procedure Button1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure OpenDialog1FilePreview(Sender: TObject;
      const Filename: string; Canvas: TCanvas; const Rect: TRect;
      Progress: TProgressEvent; var Handled: Boolean);
    procedure Button2Click(Sender: TObject);
    procedure cbPixelFormatCloseUp(Sender: TObject);
    procedure Button3Click(Sender: TObject);
  private
    { Private declarations }
    procedure OnSaveDetail(Sender: TObject; Bitmap: TCanvasOpBitmap; var Continue: Boolean);
    procedure OnProgress(Sender: TObject; Stage: TOPProgressStage;
      PercentDone: Byte; RedrawNow: Boolean; const R: TRect;
      const Msg: string);
    procedure OpPaint;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;
  fOP: TOPPicture;

implementation

uses formatdialog;

{$R *.xfm}

procedure TForm1.Button1Click(Sender: TObject);
begin
  OpenDialog1.Filter := OPGLoadFilters;
  if OpenDialog1.Execute then
  begin
    try
      fOp.LoadFromFile(OpenDialog1.Filename);
      OpPaint;
      cbPixelFormat.ItemIndex := Ord(fOP.Bitmap.PixelFormat) - 1;
      Caption := ExtractFileName(OpenDialog1.Filename) + ' (' + inttostr(fOP.Bitmap.Width) + ' x ' + inttostr(fOP.Bitmap.Height) + ')';
    except
    end;
  end;
end;

procedure TForm1.OnProgress(Sender: TObject; Stage: TOPProgressStage;
  PercentDone: Byte; RedrawNow: Boolean; const R: TRect;
  const Msg: string);
begin
  if Stage = opsEnding then ProgressBar1.Position := 100 else
    if Stage = opsStarting then ProgressBar1.Position := 0 else
      ProgressBar1.Position := PercentDone;
  Application.ProcessMessages;
end;

procedure TForm1.OnSaveDetail(Sender: TObject; Bitmap: TCanvasOpBitmap;
  var Continue: Boolean);
var FormatDialog: TFormatDialog;
begin
  FormatDialog := TFormatDialog.Create(nil);
  FormatDialog.Image := Bitmap;
  Continue := FormatDialog.ShowModal = mrOK;
  FormatDialog.free;
end;

procedure TForm1.OpPaint;
begin
  fOp.Bitmap.PixelFormat := opbitmap.pf32bit;
  Image1.Picture.Bitmap.Transparent := false;
  Image1.Transparent := false;
  Image1.Picture.Bitmap.Width := fOp.Bitmap.Width;
  Image1.Picture.Bitmap.Height := fOp.Bitmap.Height;
  Image1.Width := fOp.Bitmap.Width;
  Image1.Height := fOp.Bitmap.Height;
  Move(fOp.Bitmap.Scanline[0]^, Image1.Picture.Bitmap.Scanline[0]^, fOp.Bitmap.GetDatasize);
  if fOp.Bitmap.Transparent then
    Image1.Picture.Bitmap.TransparentColor := fOp.Bitmap.TransparentColor else
    Image1.Picture.Bitmap.Transparent := false;
  Image1.Transparent := fOp.Bitmap.Transparent;
  Image1.Invalidate;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  fOP := TOPPicture.create;
  fOP.OnProgress := OnProgress;
  fOP.OnSaveDetails := OnSaveDetail;
  Image1.Picture.Bitmap.PixelFormat := pf32bit;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  fOp.Free;
end;

procedure TForm1.OpenDialog1FilePreview(Sender: TObject;
  const Filename: string; Canvas: TCanvas; const Rect: TRect;
  Progress: TProgressEvent; var Handled: Boolean);
var Tmp: TBitmap;
begin
  fOp.OnProgress := nil;
  try
    fOp.LoadFromFile(FileName);
    fOp.Bitmap.Canvas.Resample(Rect.Right - Rect.Left, Rect.Bottom - Rect.Top);
    fOp.Bitmap.PixelFormat := opbitmap.pf32bit;
    Tmp := QGraphics.TBitmap.create;
    Tmp.PixelFormat := pf32bit;
    Tmp.Width := fOp.Bitmap.Width;
    Tmp.Height := fOp.Bitmap.Height;
    try
      Move(fOp.Bitmap.Scanline[0]^, Tmp.Scanline[0]^, fOp.Bitmap.GetDatasize);
      Canvas.Draw(0, 0, Tmp);
    finally
      Tmp.free;
    end;
    Handled := True;
  except
  end;
  fOP.OnProgress := OnProgress;
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  SaveDialog1.Filter := OPGSaveFilters;
  if SaveDialog1.Execute then
  begin
    try
      fOP.SaveToFile(SaveDialog1.FileName);
    except
    end;
  end;
end;

procedure TForm1.cbPixelFormatCloseUp(Sender: TObject);
begin
  fOP.Bitmap.PixelFormat := opbitmap.TPixelFormat(cbPixelFormat.ItemIndex + 1);
  OpPaint;
end;

procedure TForm1.Button3Click(Sender: TObject);
var Strm: TMemoryStream;
begin
  Strm := TMemoryStream.Create;
  fOp.SaveToStream(Strm, MIME_TIF);
  fOp.LoadFromStream(Strm); //Some formats like Targa need a MimeType like fOp.LoadFromStream(Strm, MIME_TGA);
  OpPaint;
  cbPixelFormat.ItemIndex := Ord(fOP.Bitmap.PixelFormat) - 1;
  Caption := fOp.MimeType + ' (' + inttostr(fOP.Bitmap.Width) + ' x ' + inttostr(fOP.Bitmap.Height) + ')';
  Strm.free;
end;

end.
